@@ -3,16 +3,17 @@ import warnings
 import threading
 
 from sqlalchemy.sql import and_, expression
-from sqlalchemy.sql.expression import ClauseElement
+from sqlalchemy.sql.expression import bindparam, ClauseElement
 from sqlalchemy.schema import Column, Index
 from sqlalchemy import func, select, false
 from sqlalchemy.schema import Table as SQLATable
 from sqlalchemy.exc import NoSuchTableError
 
 from dataset.types import Types
-from dataset.util import normalize_column_name, index_name, ensure_tuple
+from dataset.util import index_name, ensure_tuple
 from dataset.util import DatasetException, ResultIter, QUERY_STEP
-from dataset.util import normalize_table_name
+from dataset.util import normalize_table_name, pad_chunk_columns
+from dataset.util import normalize_column_name, normalize_column_key
 
 
 log = logging.getLogger(__name__)
@@ -28,9 +29,12 @@ class Table(object):
         self.db = database
         self.name = normalize_table_name(table_name)
         self._table = None
+        self._columns = None
         self._indexes = []
-        self._primary_id = primary_id
-        self._primary_type = primary_type
+        self._primary_id = primary_id if primary_id is not None \
+            else self.PRIMARY_DEFAULT
+        self._primary_type = primary_type if primary_type is not None \
+            else Types.integer
         self._auto_create = auto_create
 
     @property
@@ -47,16 +51,43 @@ class Table(object):
             self._sync_table(())
         return self._table
 
+    @property
+    def _column_keys(self):
+        """Get a dictionary of all columns and their case mapping."""
+        if not self.exists:
+            return {}
+        with self.db.lock:
+            if self._columns is None:
+                # Initialise the table if it doesn't exist
+                table = self.table
+                self._columns = {}
+                for column in table.columns:
+                    name = normalize_column_name(column.name)
+                    key = normalize_column_key(name)
+                    if key in self._columns:
+                        log.warning("Duplicate column: %s", name)
+                    self._columns[key] = name
+            return self._columns
+
+    def _flush_metadata(self):
+        with self.db.lock:
+            self._columns = None
+
     @property
     def columns(self):
         """Get a listing of all columns that exist in the table."""
-        if not self.exists:
-            return []
-        return self.table.columns.keys()
+        return list(self._column_keys.values())
 
     def has_column(self, column):
         """Check if a column with the given name exists on this table."""
-        return normalize_column_name(column) in self.columns
+        key = normalize_column_key(normalize_column_name(column))
+        return key in self._column_keys
+
+    def _get_column_name(self, name):
+        """Find the best column name with case-insensitive matching."""
+        name = normalize_column_name(name)
+        key = normalize_column_key(name)
+        return self._column_keys.get(key, name)
 
     def insert(self, row, ensure=None, types=None):
         """Add a ``row`` dict by inserting it into the table.
@@ -85,7 +116,7 @@ class Table(object):
     def insert_ignore(self, row, keys, ensure=None, types=None):
         """Add a ``row`` dict into the table if the row does not exist.
 
-        If rows with matching ``keys`` exist they will be added to the table.
+        If rows with matching ``keys`` exist no change is made.
 
         Setting ``ensure`` results in automatically creating missing columns,
         i.e., keys of the row are not table columns.
@@ -122,17 +153,29 @@ class Table(object):
             rows = [dict(name='Dolly')] * 10000
             table.insert_many(rows)
         """
-        chunk = []
+        # Sync table before inputting rows.
+        sync_row = {}
         for row in rows:
-            row = self._sync_columns(row, ensure, types=types)
+            # Only get non-existing columns.
+            sync_keys = list(sync_row.keys())
+            for key in [k for k in row.keys() if k not in sync_keys]:
+                # Get a sample of the new column(s) from the row.
+                sync_row[key] = row[key]
+        self._sync_columns(sync_row, ensure, types=types)
+
+        # Get columns name list to be used for padding later.
+        columns = sync_row.keys()
+
+        chunk = []
+        for index, row in enumerate(rows):
             chunk.append(row)
-            if len(chunk) == chunk_size:
+
+            # Insert when chunk_size is fulfilled or this is the last row
+            if len(chunk) == chunk_size or index == len(rows) - 1:
+                chunk = pad_chunk_columns(chunk, columns)
                 self.table.insert().execute(chunk)
                 chunk = []
 
-        if len(chunk):
-            self.table.insert().execute(chunk)
-
     def update(self, row, keys, ensure=None, types=None, return_count=False):
         """Update a row in the table.
 
@@ -141,7 +184,8 @@ class Table(object):
         values in ``row``.
         ::
 
-            # update all entries with id matching 10, setting their title columns
+            # update all entries with id matching 10, setting their title
+            # columns
             data = dict(id=10, title='I am a banana!')
             table.update(data, ['id'])
 
@@ -161,6 +205,44 @@ class Table(object):
         if return_count:
             return self.count(clause)
 
+    def update_many(self, rows, keys, chunk_size=1000, ensure=None,
+                    types=None):
+        """Update many rows in the table at a time.
+
+        This is significantly faster than updating them one by one. Per default
+        the rows are processed in chunks of 1000 per commit, unless you specify
+        a different ``chunk_size``.
+
+        See :py:meth:`update() <dataset.Table.update>` for details on
+        the other parameters.
+        """
+        # Convert keys to a list if not a list or tuple.
+        keys = keys if type(keys) in (list, tuple) else [keys]
+
+        chunk = []
+        columns = []
+        for index, row in enumerate(rows):
+            chunk.append(row)
+            for col in row.keys():
+                if col not in columns:
+                    columns.append(col)
+
+            # bindparam requires names to not conflict (cannot be "id" for id)
+            for key in keys:
+                row['_%s' % key] = row[key]
+
+            # Update when chunk_size is fulfilled or this is the last row
+            if len(chunk) == chunk_size or index == len(rows) - 1:
+                cl = [self.table.c[k] == bindparam('_%s' % k) for k in keys]
+                stmt = self.table.update(
+                    whereclause=and_(*cl),
+                    values={
+                        col: bindparam(col, required=False) for col in columns
+                    }
+                )
+                self.db.executable.execute(stmt, chunk)
+                chunk = []
+
     def upsert(self, row, keys, ensure=None, types=None):
         """An UPSERT is a smart combination of insert and update.
 
@@ -179,6 +261,34 @@ class Table(object):
             return self.insert(row, ensure=False)
         return True
 
+    def upsert_many(self, rows, keys, chunk_size=1000, ensure=None,
+                    types=None):
+        """
+        Sorts multiple input rows into upserts and inserts. Inserts are passed
+        to insert_many and upserts are updated.
+
+        See :py:meth:`upsert() <dataset.Table.upsert>` and
+        :py:meth:`insert_many() <dataset.Table.insert_many>`.
+        """
+        # Convert keys to a list if not a list or tuple.
+        keys = keys if type(keys) in (list, tuple) else [keys]
+
+        to_insert = []
+        to_update = []
+        for row in rows:
+            if self.find_one(**{key: row.get(key) for key in keys}):
+                # Row exists - update it.
+                to_update.append(row)
+            else:
+                # Row doesn't exist - insert it.
+                to_insert.append(row)
+
+        # Insert non-existing rows.
+        self.insert_many(to_insert, chunk_size, ensure, types)
+
+        # Update existing rows.
+        self.update_many(to_update, keys, chunk_size, ensure, types)
+
     def delete(self, *clauses, **filters):
         """Delete rows from the table.
 
@@ -200,13 +310,14 @@ class Table(object):
     def _reflect_table(self):
         """Load the tables definition from the database."""
         with self.db.lock:
+            self._flush_metadata()
             try:
                 self._table = SQLATable(self.name,
                                         self.db.metadata,
                                         schema=self.db.schema,
                                         autoload=True)
             except NoSuchTableError:
-                pass
+                self._table = None
 
     def _threading_warn(self):
         if self.db.in_transaction and threading.active_count() > 1:
@@ -217,6 +328,7 @@ class Table(object):
 
     def _sync_table(self, columns):
         """Lazy load, create or adapt the table structure in the database."""
+        self._flush_metadata()
         if self._table is None:
             # Load an existing table from the database.
             self._reflect_table()
@@ -233,8 +345,8 @@ class Table(object):
                 if self._primary_id is not False:
                     # This can go wrong on DBMS like MySQL and SQLite where
                     # tables cannot have no columns.
-                    primary_id = self._primary_id or self.PRIMARY_DEFAULT
-                    primary_type = self._primary_type or Types.integer
+                    primary_id = self._primary_id
+                    primary_type = self._primary_type
                     increment = primary_type in [Types.integer, Types.bigint]
                     column = Column(primary_id, primary_type,
                                     primary_key=True,
@@ -246,10 +358,13 @@ class Table(object):
                 self._table.create(self.db.executable, checkfirst=True)
         elif len(columns):
             with self.db.lock:
+                self._reflect_table()
                 self._threading_warn()
                 for column in columns:
                     if not self.has_column(column.name):
-                        self.db.op.add_column(self.name, column, self.db.schema)
+                        self.db.op.add_column(self.name,
+                                              column,
+                                              self.db.schema)
                 self._reflect_table()
 
     def _sync_columns(self, row, ensure, types=None):
@@ -259,23 +374,22 @@ class Table(object):
         this will remove any keys from the ``row`` for which there is no
         matching column.
         """
-        columns = self.columns
         ensure = self._check_ensure(ensure)
         types = types or {}
-        types = {normalize_column_name(k): v for (k, v) in types.items()}
+        types = {self._get_column_name(k): v for (k, v) in types.items()}
         out = {}
-        sync_columns = []
+        sync_columns = {}
         for name, value in row.items():
-            name = normalize_column_name(name)
-            if ensure and name not in columns:
+            name = self._get_column_name(name)
+            if self.has_column(name):
+                out[name] = value
+            elif ensure:
                 _type = types.get(name)
                 if _type is None:
                     _type = self.db.types.guess(value)
-                sync_columns.append(Column(name, _type))
-                columns.append(name)
-            if name in columns:
+                sync_columns[name] = Column(name, _type)
                 out[name] = value
-        self._sync_table(sync_columns)
+        self._sync_table(sync_columns.values())
         return out
 
     def _check_ensure(self, ensure):
@@ -283,15 +397,43 @@ class Table(object):
             return self.db.ensure_schema
         return ensure
 
+    def _generate_clause(self, column, op, value):
+        if op in ('like',):
+            return self.table.c[column].like(value)
+        if op in ('ilike',):
+            return self.table.c[column].ilike(value)
+        if op in ('>', 'gt'):
+            return self.table.c[column] > value
+        if op in ('<', 'lt'):
+            return self.table.c[column] < value
+        if op in ('>=', 'gte'):
+            return self.table.c[column] >= value
+        if op in ('<=', 'lte'):
+            return self.table.c[column] <= value
+        if op in ('=', '==', 'is'):
+            return self.table.c[column] == value
+        if op in ('!=', '<>', 'not'):
+            return self.table.c[column] != value
+        if op in ('in'):
+            return self.table.c[column].in_(value)
+        if op in ('between', '..'):
+            start, end = value
+            return self.table.c[column].between(start, end)
+        return false()
+
     def _args_to_clause(self, args, clauses=()):
         clauses = list(clauses)
         for column, value in args.items():
+            column = self._get_column_name(column)
             if not self.has_column(column):
                 clauses.append(false())
-            elif isinstance(value, (list, tuple)):
-                clauses.append(self.table.c[column].in_(value))
+            elif isinstance(value, (list, tuple, set)):
+                clauses.append(self._generate_clause(column, 'in', value))
+            elif isinstance(value, dict):
+                for op, op_value in value.items():
+                    clauses.append(self._generate_clause(column, op, op_value))
             else:
-                clauses.append(self.table.c[column] == value)
+                clauses.append(self._generate_clause(column, '=', value))
         return and_(*clauses)
 
     def _args_to_order_by(self, order_by):
@@ -300,7 +442,8 @@ class Table(object):
             if ordering is None:
                 continue
             column = ordering.lstrip('-')
-            if column not in self.table.columns:
+            column = self._get_column_name(column)
+            if not self.has_column(column):
                 continue
             if ordering.startswith('-'):
                 orderings.append(self.table.c[column].desc())
@@ -310,23 +453,31 @@ class Table(object):
 
     def _keys_to_args(self, row, keys):
         keys = ensure_tuple(keys)
-        keys = [normalize_column_name(k) for k in keys]
-        # keys = [self.has_column(k) for k in keys]
+        keys = [self._get_column_name(k) for k in keys]
         row = row.copy()
-        args = {k: row.pop(k) for k in keys if k in row}
+        args = {k: row.pop(k, None) for k in keys}
         return args, row
 
-    def create_column(self, name, type):
+    def create_column(self, name, type, **kwargs):
         """Create a new column ``name`` of a specified type.
         ::
 
             table.create_column('created_at', db.types.datetime)
+
+        `type` corresponds to an SQLAlchemy type as described by
+        `dataset.db.Types`. Additional keyword arguments are passed
+        to the constructor of `Column`, so that default values, and
+        options like `nullable` and `unique` can be set.
+        ::
+
+            table.create_column('key', unique=True, nullable=False)
+            table.create_column('food', default='banana')
         """
-        name = normalize_column_name(name)
+        name = self._get_column_name(name)
         if self.has_column(name):
             log.debug("Column exists: %s" % name)
             return
-        self._sync_table((Column(name, type),))
+        self._sync_table((Column(name, type, **kwargs),))
 
     def create_column_by_example(self, name, value):
         """
@@ -350,7 +501,7 @@ class Table(object):
         """
         if self.db.engine.dialect.name == 'sqlite':
             raise RuntimeError("SQLite does not support dropping columns.")
-        name = normalize_column_name(name)
+        name = self._get_column_name(name)
         with self.db.lock:
             if not self.exists or not self.has_column(name):
                 log.debug("Column does not exist: %s", name)
@@ -374,12 +525,13 @@ class Table(object):
                 self._threading_warn()
                 self.table.drop(self.db.executable, checkfirst=True)
                 self._table = None
+                self._flush_metadata()
 
     def has_index(self, columns):
         """Check if an index exists to cover the given ``columns``."""
         if not self.exists:
             return False
-        columns = set([normalize_column_name(c) for c in columns])
+        columns = set([self._get_column_name(c) for c in columns])
         if columns in self._indexes:
             return True
         for column in columns:
@@ -400,7 +552,7 @@ class Table(object):
 
             table.create_index(['name', 'country'])
         """
-        columns = [normalize_column_name(c) for c in ensure_tuple(columns)]
+        columns = [self._get_column_name(c) for c in ensure_tuple(columns)]
         with self.db.lock:
             if not self.exists:
                 raise DatasetException("Table has not been created yet.")
@@ -443,7 +595,7 @@ class Table(object):
         instead.
         """
         if not self.exists:
-            return []
+            return iter([])
 
         _limit = kwargs.pop('_limit', None)
         _offset = kwargs.pop('_offset', 0)
@@ -460,7 +612,7 @@ class Table(object):
                                   offset=_offset)
         if len(order_by):
             query = query.order_by(*order_by)
-        
+
         conn = self.db.executable
         if _streamed:
             conn = self.db.engine.connect()
@@ -521,29 +673,25 @@ class Table(object):
             table.distinct('year', country='China')
         """
         if not self.exists:
-            return []
-
-        filters = []
-        for column, value in _filter.items():
-            if not self.has_column(column):
-                raise DatasetException("No such column: %s" % column)
-            filters.append(self.table.c[column] == value)
+            return iter([])
 
         columns = []
+        clauses = []
         for column in args:
             if isinstance(column, ClauseElement):
-                filters.append(column)
+                clauses.append(column)
             else:
                 if not self.has_column(column):
                     raise DatasetException("No such column: %s" % column)
                 columns.append(self.table.c[column])
 
+        clause = self._args_to_clause(_filter, clauses=clauses)
         if not len(columns):
-            return []
+            return iter([])
 
         q = expression.select(columns,
                               distinct=True,
-                              whereclause=and_(*filters),
+                              whereclause=clause,
                               order_by=[c.asc() for c in columns])
         return self.db.query(q)
 
