@@ -1,7 +1,7 @@
-import six
 from hashlib import sha1
-from collections import OrderedDict, Sequence
-from six.moves.urllib.parse import urlparse
+from urllib.parse import urlparse
+from collections import OrderedDict
+from collections.abc import Iterable
 
 QUERY_STEP = 1000
 row_type = OrderedDict
@@ -23,7 +23,7 @@ def iter_result_proxy(rp, step=None):
         if step is None:
             chunk = rp.fetchall()
         else:
-            chunk = rp.fetchmany(step)
+            chunk = rp.fetchmany(size=step)
         if not chunk:
             break
         for row in chunk:
@@ -41,7 +41,11 @@ class ResultIter(object):
         self._iter = iter_result_proxy(result_proxy, step=step)
 
     def __next__(self):
-        return convert_row(self.row_type, next(self._iter))
+        try:
+            return convert_row(self.row_type, next(self._iter))
+        except StopIteration:
+            self.close()
+            raise
 
     next = __next__
 
@@ -54,13 +58,13 @@ class ResultIter(object):
 
 def normalize_column_name(name):
     """Check if a string is a reasonable thing to use as a column name."""
-    if not isinstance(name, six.string_types):
+    if not isinstance(name, str):
         raise ValueError('%r is not a valid column name.' % name)
 
     # limit to 63 characters
     name = name.strip()[:63]
     # column names can be 63 *bytes* max in postgresql
-    if isinstance(name, six.text_type):
+    if isinstance(name, str):
         while len(name.encode('utf-8')) >= 64:
             name = name[:len(name) - 1]
 
@@ -69,9 +73,16 @@ def normalize_column_name(name):
     return name
 
 
+def normalize_column_key(name):
+    """Return a comparable column name."""
+    if name is None or not isinstance(name, str):
+        return None
+    return name.upper().strip().replace(' ', '')
+
+
 def normalize_table_name(name):
     """Check if the table name is obviously invalid."""
-    if not isinstance(name, six.string_types):
+    if not isinstance(name, str):
         raise ValueError("Invalid table name: %r" % name)
     name = name.strip()[:63]
     if not len(name):
@@ -99,6 +110,15 @@ def ensure_tuple(obj):
     """Try and make the given argument into a tuple."""
     if obj is None:
         return tuple()
-    if isinstance(obj, Sequence) and not isinstance(obj, six.string_types):
+    if isinstance(obj, Iterable) and not isinstance(obj, (str, bytes)):
         return tuple(obj)
     return obj,
+
+
+def pad_chunk_columns(chunk, columns):
+    """Given a set of items to be inserted, make sure they all have the
+    same columns by padding columns with None if they are missing."""
+    for record in chunk:
+        for column in columns:
+            record.setdefault(column, None)
+    return chunk
