@@ -9,7 +9,7 @@ from datetime import datetime
 from sqlalchemy import FLOAT, TEXT, BIGINT
 from sqlalchemy.exc import IntegrityError, SQLAlchemyError, ArgumentError
 
-from dataset import connect
+from dataset import connect, chunked
 
 from .sample_data import TEST_DATA, TEST_CITY_1
 
@@ -394,6 +394,26 @@ class TableTestCase(unittest.TestCase):
         self.tbl.insert_many(data, chunk_size=13)
         assert len(self.tbl) == len(data) + 6
 
+    def test_chunked_insert(self):
+        data = TEST_DATA * 100
+        with chunked.ChunkedInsert(self.tbl) as chunk_tbl:
+            for item in data:
+                chunk_tbl.insert(item)
+        assert len(self.tbl) == len(data) + 6
+
+    def test_chunked_insert_callback(self):
+        data = TEST_DATA * 100
+        N = 0
+
+        def callback(queue):
+            nonlocal N
+            N += len(queue)
+        with chunked.ChunkedInsert(self.tbl, callback=callback) as chunk_tbl:
+            for item in data:
+                chunk_tbl.insert(item)
+        assert len(data) == N
+        assert len(self.tbl) == len(data) + 6
+
     def test_update_many(self):
         tbl = self.db['update_many_test']
         tbl.insert_many([
@@ -406,6 +426,22 @@ class TableTestCase(unittest.TestCase):
         # Ensure data has been updated.
         assert tbl.find_one(id=1)['temp'] == tbl.find_one(id=3)['temp']
 
+    def test_chunked_update(self):
+        tbl = self.db['update_many_test']
+        tbl.insert_many([
+            dict(temp=10, location='asdf'), dict(temp=20, location='qwer'), dict(temp=30, location='asdf')
+        ])
+
+        chunked_tbl = chunked.ChunkedUpdate(tbl, 'id')
+        chunked_tbl.update(dict(id=1, temp=50))
+        chunked_tbl.update(dict(id=2, location='asdf'))
+        chunked_tbl.update(dict(id=3, temp=50))
+        chunked_tbl.flush()
+
+        # Ensure data has been updated.
+        assert tbl.find_one(id=1)['temp'] == tbl.find_one(id=3)['temp'] == 50
+        assert tbl.find_one(id=2)['location'] == tbl.find_one(id=3)['location'] == 'asdf'
+
     def test_upsert_many(self):
         # Also tests updating on records with different attributes
         tbl = self.db['upsert_many_test']
